#include "nvfs.h"

static inline ssize_t nvfs_read_locked(struct nvfs_memory_inode *nmi, struct file *file, char __user *buf, size_t len, loff_t *ppos)
{
	struct nvfs_superblock *nvs = nmi->vfs_inode.i_sb->s_fs_info;
	ssize_t total = 0;
	loff_t pos;
	int r;
	size_t i_size;

	if (unlikely(!len))
		return 0;

	pos = *ppos;

	i_size = nmi->vfs_inode.i_size;
	if (pos >= i_size)
		return 0;
	if (pos + len > i_size)
		len = i_size - pos;

	if (unlikely(!access_ok(buf, len)))
		return -EFAULT;

	do {
		void *blk;

		unsigned page_offset = pos & ((1UL << nvs->log2_page_size) - 1);
		unsigned prealloc = (len + (1UL << nvs->log2_page_size) - 1) >> nvs->log2_page_size;
		unsigned size;

		blk = nvfs_bmap(nmi, pos >> nvs->log2_page_size, &prealloc, NULL, NULL, NULL);
		if (unlikely(IS_ERR(blk))) {
			r = PTR_ERR(blk);
			goto ret_r;
		}
		size = min(len, ((size_t)prealloc << nvs->log2_page_size) - page_offset);
		if (unlikely(!blk)) {
			size = min(size, (unsigned)PAGE_SIZE);
			size -= __copy_to_user(buf, empty_zero_page, size);
		} else {
			size -= __copy_to_user(buf, blk + page_offset, size);
		}

		if (unlikely(!size)) {
			r = -EFAULT;
			goto ret_r;
		}

		buf += size;
		len -= size;
		pos += size;
		total += size;
	} while (len);

	r = 0;

ret_r:
	*ppos = pos;

	if (file)
		file_accessed(file);

	return total ? total : r;
}

struct nvfs_aux {
	const char __user *user_buf;
	size_t size;
	unsigned page_offset;
	unsigned copied;
	bool clear;
};

#if 0
static inline void cache_line_copy(void *dest, const void *src, size_t len)
{
	stac();
	if (unlikely(((unsigned long)dest & 63) != 0)) {
		size_t len1 = -(unsigned long)dest & 63;
		if (unlikely(len1 > len))
			len1 = len;
		memcpy(dest, src, len1);
		clwb(dest);
		dest += len1;
		src += len1;
		len -= len1;
	}
	while (len >= 64) {
		memcpy(dest, src, 64);
		clwb(dest);
		dest += 64;
		src += 64;
		len -= 64;
	}
	if (unlikely(len != 0)) {
		memcpy(dest, src, len);
		clwb(dest);
	}
	clac();
}
#endif

#if defined(CONFIG_X86_64)
long __copy_user_flushcache(void *dst, const void __user *src, unsigned size)
{
	unsigned long flushed, dest = (unsigned long) dst;
	long rc = __copy_user_nocache(dst, src, size, 0);

	/*
	 * __copy_user_nocache() uses non-temporal stores for the bulk
	 * of the transfer, but we need to manually flush if the
	 * transfer is unaligned. A cached memory copy is used when
	 * destination or size is not naturally aligned. That is:
	 *   - Require 8-byte alignment when size is 8 bytes or larger.
	 *   - Require 4-byte alignment when size is 4 bytes.
	 */
	if (size < 8) {
		if (!IS_ALIGNED(dest, 4) || size != 4)
			arch_wb_cache_pmem(dst, size);
	} else {
		if (!IS_ALIGNED(dest, 8)) {
			dest = ALIGN(dest, boot_cpu_data.x86_clflush_size);
			arch_wb_cache_pmem(dst, 1);
		}

		flushed = dest - (unsigned long) dst;
		if (size > flushed && !IS_ALIGNED(size - flushed, 8))
			arch_wb_cache_pmem(dst + size - 1, 1);
	}

	return rc;
}
#endif

static void nvfs_write_page(struct nvfs_superblock *nvs, void *page, unsigned prealloc, void *aux_)
{
	struct nvfs_aux *aux = aux_;
	unsigned total_size = (prealloc << nvs->log2_page_size) - aux->page_offset;
	if (total_size > aux->size)
		total_size = aux->size;

	if (aux->clear)
		nvfs_memzero_flushcache(page, aux->page_offset);

#if defined(CONFIG_X86_64)
	aux->copied = total_size - __copy_user_flushcache(page + aux->page_offset, aux->user_buf, total_size);
#elif 1
	/* this is better for multi-threaded workload - 4% improvement */
	aux->copied = total_size - __copy_from_user(page + aux->page_offset, aux->user_buf, total_size);
	arch_wb_cache_pmem(page + aux->page_offset, aux->copied);
#else
	/* the performance benefit of this is small - about 2.5% */
	cache_line_copy(page + aux->page_offset, aux->user_buf, total_size);
	aux->copied = total_size;
#endif

	if (aux->clear)
		nvfs_memzero_flushcache(page + aux->page_offset + aux->copied, (prealloc << nvs->log2_page_size) - (aux->page_offset + aux->copied));
}

static inline ssize_t nvfs_write_locked(struct nvfs_memory_inode *nmi, struct file *file, const char __user *buf, size_t len, loff_t *ppos)
{
	struct nvfs_superblock *nvs = nmi->vfs_inode.i_sb->s_fs_info;
	ssize_t total = 0;
	loff_t start_pos, pos;
	int r;
	bool some_new = false;

	if (unlikely(!len))
		return 0;

	if (unlikely(!access_ok(buf, len)))
		return -EFAULT;

	if (file) {
		r = file_modified(file);
		if (unlikely(r < 0))
			return r;
	}

	if (ppos)
		pos = *ppos;
	else
		pos = nmi->vfs_inode.i_size;
	start_pos = pos;

	if (unlikely((uint64_t)pos + len < len) ||
	    unlikely((uint64_t)pos + len > (uint64_t)nvs->s->s_maxbytes)) {
		return -EFBIG;
	}

	while (len) {
		struct nvfs_aux aux;
		void *blk;
		unsigned n_pages;
		bool new;

		aux.user_buf = buf;
		aux.page_offset = pos & ((1UL << nvs->log2_page_size) - 1);
		aux.size = len;
		aux.clear = true;
		n_pages = (aux.page_offset + len + (1UL << nvs->log2_page_size) - 1) >> nvs->log2_page_size;

		blk = nvfs_bmap(nmi, pos >> nvs->log2_page_size, &n_pages, nvfs_write_page, &aux, &new);
		if (unlikely(IS_ERR(blk))) {
			r = PTR_ERR(blk);
			goto ret_r;
		}
		if (unlikely(!blk)) {
			r = -ENOSPC;
			goto ret_r;
		}
		some_new |= new;
		if (!new) {
			aux.clear = false;
			nvfs_write_page(nvs, blk, 1, &aux);
		}

		if (unlikely(!aux.copied)) {
			r = -EFAULT;
			goto ret_r;
		}

		buf += aux.copied;
		len -= aux.copied;
		pos += aux.copied;
		total += aux.copied;
	}

	r = 0;

ret_r:
	if (some_new && unlikely(nmi->was_mmapped)) {
		if (pos > start_pos) {
			uint64_t start_fs_page, end_fs_page, start_page, end_page;

			start_fs_page = start_pos >> nvs->log2_page_size;
			end_fs_page = (pos + (1UL << nvs->log2_page_size) - 1) >> nvs->log2_page_size;

			start_page = start_fs_page << (nvs->log2_page_size - PAGE_SHIFT);
			end_page = end_fs_page << (nvs->log2_page_size - PAGE_SHIFT);

			invalidate_inode_pages2_range(nmi->vfs_inode.i_mapping, start_page, end_page - 1);
		}
	}
	if (pos > nmi->vfs_inode.i_size) {
		pmem_wmb();
		nvfs_set_inode_size(nmi, pos);
	}

	if (ppos)
		*ppos = pos;

	return total ? total : r;
}

static ssize_t nvfs_read(struct file *file, char __user *buf, size_t len, loff_t *ppos)
{
	ssize_t r;
	struct nvfs_memory_inode *nmi = i_to_nmi(file->f_inode);
	inode_lock_shared(&nmi->vfs_inode);
	if (!IS_DAX(&nmi->vfs_inode))
		r = nvfs_cached_rw_locked(nmi, file, false, buf, len, ppos);
	else
		r = nvfs_read_locked(nmi, file, buf, len, ppos);
	inode_unlock_shared(&nmi->vfs_inode);
	return r;
}

static ssize_t nvfs_write(struct file *file, const char __user *buf, size_t len, loff_t *ppos)
{
	ssize_t r;
	struct nvfs_memory_inode *nmi = i_to_nmi(file->f_inode);
	bool shared_lock;
	if (unlikely((file->f_flags & (O_APPEND | O_DIRECT)) == O_DIRECT)) {
		inode_lock_shared(&nmi->vfs_inode);
		if (unlikely(*ppos + len > nmi->vfs_inode.i_size)) {
			inode_unlock_shared(&nmi->vfs_inode);
			goto ex_lock;
		}
		shared_lock = true;
	} else {
ex_lock:
		inode_lock(&nmi->vfs_inode);
		shared_lock = false;
	}
	if (file->f_flags & O_APPEND)
		ppos = NULL;
	if (!IS_DAX(&nmi->vfs_inode))
		r = nvfs_cached_rw_locked(nmi, file, true, (char __user *)buf, len, ppos);
	else
		r = nvfs_write_locked(nmi, file, buf, len, ppos);
	if (unlikely(shared_lock))
		inode_unlock_shared(&nmi->vfs_inode);
	else
		inode_unlock(&nmi->vfs_inode);
	return r;
}

ssize_t nvfs_rw_iter_locked(struct kiocb *iocb, struct iov_iter *iov)
{
	ssize_t r = 0;
	ssize_t total = 0;
	bool wr = iov->type & WRITE;
	struct nvfs_memory_inode *nmi = i_to_nmi(iocb->ki_filp->f_inode);

	while (iov_iter_count(iov)) {
		struct iovec iovec;
		mm_segment_t old_fs = get_fs();
		switch (iov_iter_type(iov)) {
			case ITER_IOVEC:
				iovec = iov_iter_iovec(iov);
				break;
			case ITER_BVEC:
				set_fs(KERNEL_DS);
				iovec.iov_base = kmap(iov->bvec->bv_page) + iov->bvec->bv_offset;
				iovec.iov_len = iov->bvec->bv_len;
				break;
			case ITER_KVEC:
				set_fs(KERNEL_DS);
				iovec.iov_base = iov->kvec->iov_base;
				iovec.iov_len = iov->kvec->iov_len;
				break;
			default:
				return -EINVAL;
		}

		if (!wr) {
			r = nvfs_read_locked(nmi, iocb->ki_filp, iovec.iov_base, iovec.iov_len, &iocb->ki_pos);
		} else {
			loff_t *ppos = &iocb->ki_pos;
			if (iocb->ki_flags & IOCB_APPEND)
				ppos = NULL;
			r = nvfs_write_locked(nmi, iocb->ki_filp, iovec.iov_base, iovec.iov_len, ppos);
		}
		switch (iov_iter_type(iov)) {
			case ITER_BVEC:
				kunmap(iov->bvec->bv_page);
				/* fallthrough */
			case ITER_KVEC:
				set_fs(old_fs);
				break;
			default:
				break;
		}
		if (unlikely(r <= 0))
			break;

		iov_iter_advance(iov, r);

		total += r;
		if (unlikely(r != iovec.iov_len))
			break;
	}

	return total ? total : r;
}

static ssize_t nvfs_rw_iter(struct kiocb *iocb, struct iov_iter *iov)
{
	ssize_t r = 0;
	bool wr = iov->type & WRITE;
	struct nvfs_memory_inode *nmi = i_to_nmi(iocb->ki_filp->f_inode);
	bool shared_lock;
	if (!wr) {
		inode_lock_shared(&nmi->vfs_inode);
		shared_lock = true;
	} else if (unlikely((iocb->ki_flags & (IOCB_DIRECT | IOCB_APPEND)) == IOCB_DIRECT)) {
		inode_lock_shared(&nmi->vfs_inode);
		if (unlikely(iocb->ki_pos + iov_iter_count(iov) > nmi->vfs_inode.i_size)) {
			inode_unlock_shared(&nmi->vfs_inode);
			goto ex_lock;
		}
		shared_lock = true;
	} else {
ex_lock:
		inode_lock(&nmi->vfs_inode);
		shared_lock = false;
	}

	if (!IS_DAX(&nmi->vfs_inode)) {
		//iocb->ki_flags &= ~IOCB_DIRECT;
		if (!wr)
			r = generic_file_read_iter(iocb, iov);
		else
			r = __generic_file_write_iter(iocb, iov);
		goto brk;
	}

	r = nvfs_rw_iter_locked(iocb, iov);

brk:
	if (unlikely(shared_lock))
		inode_unlock_shared(&nmi->vfs_inode);
	else
		inode_unlock(&nmi->vfs_inode);

	return r;
}

int nvfs_fsync_file(struct file *file, loff_t start, loff_t end, int datasync)
{
	struct inode *inode = file_inode(file);
	if (!IS_DAX(inode))
		return __generic_file_fsync(file, start, end, datasync);
	else
		return 0;
}

static long nvfs_fallocate(struct file *file, int mode, loff_t offset, loff_t len)
{
	struct inode *inode = file_inode(file);
	int r;

	if (mode & FALLOC_FL_PUNCH_HOLE) {
		inode_lock(inode);
		r = nvfs_punch_hole(i_to_nmi(inode), offset, offset + len);
		inode_unlock(inode);

		return r;
	}

	return -EOPNOTSUPP;
}

#ifdef CONFIG_QUOTA

ssize_t nvfs_quota_read(struct super_block *sb, int type, char *buf, size_t len, loff_t off)
{
	struct nvfs_memory_inode *nmi = i_to_nmi(sb_dqopt(sb)->files[type]);
	ssize_t r;
	mm_segment_t oldfs = get_fs();
	set_fs(KERNEL_DS);

	r = nvfs_read_locked(nmi, NULL, buf, len, &off);

	set_fs(oldfs);
	return r;
}

ssize_t nvfs_quota_write(struct super_block *sb, int type, const char *buf, size_t len, loff_t off)
{
	struct nvfs_memory_inode *nmi = i_to_nmi(sb_dqopt(sb)->files[type]);
	ssize_t r;
	mm_segment_t oldfs = get_fs();
	set_fs(KERNEL_DS);

	r = nvfs_write_locked(nmi, NULL, buf, len, &off);

	set_fs(oldfs);
	return r;
}

#endif

const struct inode_operations nvfs_file_inode_operations = {
	.getattr = nvfs_getattr,
	.setattr = nvfs_setattr,
	.update_time = nvfs_update_time,
	.listxattr = nvfs_listxattr,
	.get_acl = nvfs_get_acl,
	.set_acl = nvfs_set_acl,
};

const struct file_operations nvfs_file_fops = {
	.llseek = generic_file_llseek,
	.read = nvfs_read,
	.write = nvfs_write,
	.read_iter = nvfs_rw_iter,
	.write_iter = nvfs_rw_iter,
	.fsync = nvfs_fsync_file,
	.fallocate = nvfs_fallocate,
	.mmap = nvfs_mmap,
	.unlocked_ioctl = nvfs_ioctl,
#ifdef CONFIG_COMPAT
	.compat_ioctl = nvfs_compat_ioctl,
#endif
};
