#include "fsck.h"

static int processing_lost_inodes = 0;

static void scan_directory(uint64_t ino);

static bool validate_ino(uint64_t ino)
{
	if (ino >= n_groups << log2_inodes_per_group)
		return false;
	if (is_inode_reserved(ino))
		return false;
	return true;
}

#define FT_UNKNOWN	0
#define FT_REG_FILE	1
#define FT_DIR		2
#define FT_CHRDEV	3
#define FT_BLKDEV	4
#define FT_FIFO		5
#define FT_SOCK		6
#define FT_SYMLINK	7

static unsigned char mode2ftype(mode_t mode)
{
	switch (mode & S_IFMT) {
		case S_IFIFO:
			return FT_FIFO + 1;
		case S_IFCHR:
			return FT_CHRDEV + 1;
		case S_IFDIR:
			return FT_DIR + 1;
		case S_IFBLK:
			return FT_BLKDEV + 1;
		case S_IFREG:
			return FT_REG_FILE + 1;
		case S_IFLNK:
			return FT_SYMLINK + 1;
		case S_IFSOCK:
			return FT_SOCK + 1;
		default:
			return 0;
	}
}

static void scan_directory_page(uint64_t ino, struct nvfs_dir_head *ndh)
{
	struct nvfs_dir_entry *nde, *end;
	uint64_t next;

	uint64_t avoid_loop_1 = 0, avoid_loop_2 = 0;

again:
	nde = ndh->nde;
	end = (void *)ndh + (1U << sb->log2_page_size);

	do {
		uint32_t namelen = nvfs_dir_entry_namelen(nde);
		size_t size = nvfs_dir_entry_needed_size(namelen);
		unsigned char type = nvfs_dir_entry_type(nde);
		if ((size_t)((void *)end - (void *)nde) < size) {
			return;
		}
		if (type) {
			struct nvfs_inode *sub_inode;
			uint64_t sub_ino = le64_to_cpu(nde->inode);
			unsigned char ftype;
			//fprintf(stderr, "testing '%.*s'\n", (int)namelen, nde->name);
			if (!validate_ino(sub_ino)) {
				static struct qkey q;
				if (fsck_query(&q, Q_NORMAL, "a directory entry '%.*s' points to an invalid inode", "delete?", (int)namelen, nde->name)) {
del_dir_entry:
					nde->name_len_and_type &= cpu_to_le32(NAME_LEN_MASK);
				}
				goto next;
			}

			if (processing_lost_inodes) {
				if (bmp_op(sub_ino, B_READ | B_INODE | B_DEV) && !bmp_op(sub_ino, B_READ | B_INODE | B_MEM))
					set_ino_flag(sub_ino);
				goto next;
			}

			if (!bmp_op(sub_ino, B_READ | B_INODE | B_DEV)) {
				static struct qkey q;
				if (fsck_query(&q, Q_NORMAL, "a directory entry '%.*s' points to a free inode", "delete?", (int)namelen, nde->name)) {
					goto del_dir_entry;
				}
				goto next;
			}
			sub_inode = p_inode(sub_ino);

			ftype = mode2ftype(le32_to_cpu(sub_inode->i_mode));
				/* this can happen with "-fn" because we are not fixing inodes */
			if (!ftype)
				goto next;

			if (type != ftype) {
				static struct qkey q;
				if (fsck_query(&q, Q_PREEN, "invalid dirent type: %x != %x", "fix?", (unsigned)type, (unsigned)ftype)) {
					uint32_t nl = le32_to_cpu(nde->name_len_and_type);
					nl &= ~TYPE_MASK;
					nl |= ftype << TYPE_OFFSET;
					nde->name_len_and_type = cpu_to_le32(nl);
				}
			}

			if (S_ISDIR(le32_to_cpu(sub_inode->i_mode))) {
				if (!bmp_op(sub_ino, B_SET | B_INODE | B_MEM)) {
					static struct qkey q;
					if (fsck_query(&q, Q_NORMAL, "a directory entry '%.*s' points to a cross-linked directory", "delete?", (int)namelen, nde->name)) {
						goto del_dir_entry;
					}
					goto next;
				}
				if (!inc_n_links(ino)) {
					static struct qkey q;
					if (fsck_query(&q, Q_NORMAL, "a directory %"PRIu64" has too many subdirectories", "delete?", ino)) {
						goto del_dir_entry;
					}
				}
#pragma omp task
				scan_directory(sub_ino);
			} else {
				if (!inc_n_links(sub_ino)) {
					static struct qkey q;
					if (fsck_query(&q, Q_NORMAL, "inode %"PRIu64" has too many links", "unlink?", sub_ino)) {
						goto del_dir_entry;
					}
					goto next;
				}
			}
		}
next:
		nde = (void *)nde + size;
	} while (nde < end);

	next = le64_to_cpu(ndh->next);
	if (next) {
		if (!verify_block_range(next))
			return;
		if (next == avoid_loop_1)
			return;
		avoid_loop_2++;
		if (!(avoid_loop_2 & (avoid_loop_2 - 1)))
			avoid_loop_1 = next;
		ndh = ptr + next;
		goto again;
	}
}

static void scan_directory_run(uint64_t ino, __le64 *run, unsigned len, unsigned depth)
{
	unsigned i;
	if (depth >= 5) {
		return;
	}
	for (i = 0; i < len; i++) {
		if (!i || run[i - 1] != run[i]) {
			uint64_t p = le64_to_cpu(run[i]);
			if (!verify_block_range(p))
				continue;
			if (*(__le64 *)(ptr + p) == NVFS_DIR_HEAD_MAGIC) {
				scan_directory_page(ino, ptr + p);
			} else {
				scan_directory_run(ino, (__le64 *)(ptr + p), 1U << (sb->log2_page_size - 3), depth + 1);
			}
		}
	}
}

uint64_t processed_dirs = 0;

static void scan_directory(uint64_t ino)
{
	struct nvfs_inode *inode = p_inode(ino);
	__le64 *run;

#pragma omp atomic
	processed_dirs++;

	if (!(processed_dirs % DIR_PROGRESS_STEP))
		fsck_progress(3, n_dirs, processed_dirs);

	run = inode->i_radix[(le32_to_cpu(inode->i_flag) / NVFS_FLAG_SECOND_DIR_BANK) & 1];
	//volatile int i; for (i = 0; i < 100000; i++) ;

	if (!processing_lost_inodes) {
		if (!inc_n_links(ino))
			fsck_internal_error("scan_directory: could not create initial reference");
	}

	scan_directory_run(ino, run, NVFS_RADIX_BLOCKS, 0);
}

static uint64_t init_directory(void)
{
	struct nvfs_inode *inode;
	uint64_t ino, b;
	unsigned i;
	struct timespec ts;

	ino = alloc_block_or_inode(B_INODE);
	if (!ino)
		return false;

	b = alloc_dir_leaf(ino);
	if (!b) {
		bmp_op(ino, B_CLEAR | B_INODE | B_MEM);
		bmp_op(ino, B_CLEAR | B_INODE | B_DEV);
		return false;
	}

	inode = p_inode(ino);
	memset(inode, 0, 1U << sb->log2_inode_size);

	inode->i_size = cpu_to_le64(1U << sb->log2_page_size);
	inode->i_blocks = cpu_to_le64(1U << (sb->log2_page_size - 9));
	for (i = 0; i < NVFS_RADIX_BLOCKS; i++)
		inode->i_radix[0][i] = cpu_to_le64(b);
	if (clock_gettime(CLOCK_REALTIME, &ts)) {
		fsck_log("clock_gettime failed");
		return 0;
	}
	inode->i_ctime = inode->i_mtime = inode->i_atime = inode->i_btime = cpu_to_le64(ts.tv_sec);
	inode->i_ctime_nsec = inode->i_mtime_nsec = inode->i_atime_nsec = inode->i_btime_nsec = cpu_to_le32(ts.tv_nsec);
	inode->i_links = cpu_to_le32(2);
	inode->i_mode = cpu_to_le32(S_IFDIR | 0755);
	inode->i_xattr[0].x = nvfs_xattr_entry_make_size((1U << sb->log2_inode_size) - offsetof(struct nvfs_inode, i_xattr));

	return ino;
}

static void free_empty_directory(uint64_t ino)
{
	struct nvfs_inode *inode;
	uint64_t block;

	inode = p_inode(ino);

	block = le64_to_cpu(inode->i_radix[0][0]) >> sb->log2_page_size;
	bmp_op(block, B_CLEAR | B_BLOCK | B_MEM);
	bmp_op(block, B_CLEAR | B_BLOCK | B_DEV);

	bmp_op(ino, B_CLEAR | B_INODE | B_MEM);
	bmp_op(ino, B_CLEAR | B_INODE | B_DEV);
}

static uint64_t init_lost_plus_found(void)
{
	char name[128];
	unsigned long long seq = 1;
	uint64_t lpf_ino, root_ino;
	struct nvfs_dir_entry *nde;

	root_ino = le64_to_cpu(sb->root_inode);
	lpf_ino = init_directory();
	if (!lpf_ino)
		goto err_0;

next_seq:
	snprintf(name, sizeof name, "lost+found.%llu", seq);

	if (nvfs_find_de(root_ino, name, strlen(name))) {
		seq++;
		goto next_seq;
	}

	if (!(nde = add_directory_entry(root_ino, name, strlen(name), lpf_ino, mode2ftype(S_IFDIR))))
		goto err_1;

	if (!inc_n_links(root_ino))
		goto err_2;
	if (!inc_n_links(lpf_ino))
		fsck_internal_error("init_lost_plus_found: could not increment n_links on the lost+found directory");
	p_inode(root_ino)->i_links = cpu_to_le32(le32_to_cpu(p_inode(root_ino)->i_links) + 1);

	{
		static struct qkey q;
		fsck_query(&q, Q_JUST_PRINT, "created the directory '%s'", NULL, name);
	}

	return lpf_ino;

err_2:
	nde->name_len_and_type &= cpu_to_le32(NAME_LEN_MASK);
err_1:
	free_empty_directory(lpf_ino);
err_0:
	{
		static struct qkey q;
		fsck_query(&q, Q_JUST_PRINT, "could not create the lost+found directory", NULL);
	}
	return 0;
}

static bool add_to_lost_plus_found(uint64_t lpf_ino, uint64_t ino)
{
	unsigned long long seq = 0;
	char name[128];
	struct nvfs_inode *inode;
	struct nvfs_dir_entry *nde;

next_seq:
	if (!seq)
		snprintf(name, sizeof name, "%"PRIu64"", ino);
	else
		snprintf(name, sizeof name, "%"PRIu64".%llu", ino, seq);

	if (nvfs_find_de(lpf_ino, name, strlen(name))) {
		seq++;
		goto next_seq;
	}

	{
		static struct qkey q;
		fsck_query(&q, Q_JUST_PRINT, "adding file %s to lost+found", NULL, name);
	}

	inode = p_inode(ino);

	if (!(nde = add_directory_entry(lpf_ino, name, strlen(name), ino, mode2ftype(le32_to_cpu(inode->i_mode)))))
		goto err_0;
	if (S_ISDIR(le32_to_cpu(inode->i_mode))) {
		struct nvfs_inode *inode = p_inode(lpf_ino);
		if (!inc_n_links(lpf_ino))
			goto err_1;
		inode->i_links = cpu_to_le32(le32_to_cpu(inode->i_links) + 1);
	}

	if (!inc_n_links(ino))
		goto err_2;

	return true;

err_2:
	if (S_ISDIR(le32_to_cpu(inode->i_mode))) {
		dec_n_links(lpf_ino);
	}
err_1:
	nde->name_len_and_type &= cpu_to_le32(NAME_LEN_MASK);
err_0:
	{
		static struct qkey q;
		fsck_query(&q, Q_JUST_PRINT, "could not add the file '%s' to the lost+found directory", NULL, name);
	}
	return false;
}

static int compare_inodes(const void *p1, const void *p2)
{
	uint64_t ino1 = *(const uint64_t *)p1;
	uint64_t ino2 = *(const uint64_t *)p2;
	struct nvfs_inode *inode1 = p_inode(ino1);
	struct nvfs_inode *inode2 = p_inode(ino2);
	if (S_ISDIR(le32_to_cpu(inode1->i_mode)) && !S_ISDIR(le32_to_cpu(inode2->i_mode)))
		return -1;
	if (!S_ISDIR(le32_to_cpu(inode1->i_mode)) && S_ISDIR(le32_to_cpu(inode2->i_mode)))
		return 1;
	if (!test_ino_flag(ino1) && test_ino_flag(ino2))
		return -1;
	if (test_ino_flag(ino1) && !test_ino_flag(ino2))
		return 1;
	if (ino1 < ino2)
		return -1;
	if (ino1 > ino2)
		return 1;
	return 0;
}

int check_connectivity(void)
{
	uint64_t ino;
	uint64_t lpf_ino = 0;

	uint64_t *li;
	size_t n_li;

	size_t i;

	fsck_log("checking directory connectivity...");

	ino = le64_to_cpu(sb->root_inode);
	if (!validate_ino(ino) ||
	    !bmp_op(ino, B_READ | B_INODE | B_DEV) ||
	    !S_ISDIR(le32_to_cpu(p_inode(ino)->i_mode))) {
		static struct qkey q;
		if (!fsck_query(&q, Q_NORMAL, "invalid root directory inode %"PRIu64"", "make new?", ino))
			return 4;
		ino = init_directory();
		if (!ino) {
			static struct qkey q;
			fsck_query(&q, Q_JUST_PRINT, "failed to allocate root inode", NULL);
			return 4;
		}
		sb->root_inode = cpu_to_le64(ino);
		superblock_modified();
		goto scan_tree;
	}

	if (!bmp_op(ino, B_SET | B_INODE | B_MEM)) {
		fsck_internal_error("check_connectivity: the root inode %"PRIu64" is already marked as used", ino);
	}
scan_tree:

#pragma omp parallel
#pragma omp single
#pragma omp task
	scan_directory(ino);

	init_array(&li, &n_li);

	for (ino = 0; ino < n_groups << log2_inodes_per_group; ino++) {
		if (!bmp_op(ino, B_READ | B_INODE | B_DEV))
			continue;
		if (bmp_op(ino, B_READ | B_INODE | B_MEM))
			continue;
		if (!lpf_ino && !ro) {
			static struct qkey q;
			if (fsck_query(&q, Q_PREEN, "there are lost inodes", "create lost+found?")) {
				lpf_ino = init_lost_plus_found();
			}
			if (!lpf_ino)
				lpf_ino = le64_to_cpu(sb->root_inode);
		}
		append_array(&li, &n_li, ino);
	}

	if (!lpf_ino)
		goto ret;

	fsck_log("processing lost inodes...");

	processing_lost_inodes = 1;
	for (i = 0; i < n_li; i++) {
		uint64_t ino = li[i];
		struct nvfs_inode *inode = p_inode(ino);
		if (S_ISDIR(le32_to_cpu(inode->i_mode)))
			scan_directory(ino);
	}
	processing_lost_inodes = 0;

	qsort(li, n_li, sizeof(uint64_t), compare_inodes);

	for (i = 0; i < n_li; i++) {
		clear_ino_flag(li[i]);
	}

	for (i = 0; i < n_li; i++) {
		static struct qkey q;
		uint64_t ino = li[i];
		struct nvfs_inode *inode = p_inode(ino);
		if (!bmp_op(ino, B_READ | B_INODE | B_DEV))
			continue;
		if (bmp_op(ino, B_READ | B_INODE | B_MEM))
			continue;
		if (!fsck_query(&q, Q_PREEN, "lost inode %"PRIu64"", "add to lost+found?", ino))
			continue;
		if (!add_to_lost_plus_found(lpf_ino, ino))
			continue;
		if (S_ISDIR(le32_to_cpu(inode->i_mode)))
			scan_directory(ino);
	}

ret:
	fsck_free(li);

	return 0;
}
